@@ -30,7 +30,7 @@
 #include "mm-modem-helpers-mbim.h"
 #include "mm-port-enums-types.h"
 #include "mm-bearer-mbim.h"
-#include "mm-log.h"
+#include "mm-log-object.h"
 
 G_DEFINE_TYPE (MMBearerMbim, mm_bearer_mbim, MM_TYPE_BASE_BEARER)
 
@@ -211,6 +211,7 @@ typedef enum {
     CONNECT_STEP_FIRST,
     CONNECT_STEP_PACKET_SERVICE,
     CONNECT_STEP_PROVISIONED_CONTEXTS,
+    CONNECT_STEP_CHECK_DISCONNECTED,
     CONNECT_STEP_ENSURE_DISCONNECTED,
     CONNECT_STEP_CONNECT,
     CONNECT_STEP_IP_CONFIGURATION,
@@ -252,6 +253,7 @@ ip_configuration_query_ready (MbimDevice *device,
                               GAsyncResult *res,
                               GTask *task)
 {
+    MMBearerMbim *self;
     ConnectContext *ctx;
     GError *error = NULL;
     MbimMessage *response;
@@ -270,7 +272,8 @@ ip_configuration_query_ready (MbimDevice *device,
     guint32 ipv4mtu;
     guint32 ipv6mtu;
 
-    ctx = g_task_get_task_data (task);
+    self = g_task_get_source_object (task);
+    ctx  = g_task_get_task_data (task);
 
     response = mbim_device_command_finish (device, res, &error);
     if (response &&
@@ -301,17 +304,17 @@ ip_configuration_query_ready (MbimDevice *device,
         /* IPv4 info */
 
         str = mbim_ip_configuration_available_flag_build_string_from_mask (ipv4configurationavailable);
-        mm_dbg ("IPv4 configuration available: '%s'", str);
+        mm_obj_dbg (self, "IPv4 configuration available: '%s'", str);
         g_free (str);
 
         if ((ipv4configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_ADDRESS) && ipv4addresscount) {
             guint i;
 
-            mm_dbg ("  IP addresses (%u)", ipv4addresscount);
+            mm_obj_dbg (self, "  IP addresses (%u)", ipv4addresscount);
             for (i = 0; i < ipv4addresscount; i++) {
                 addr = g_inet_address_new_from_bytes ((guint8 *)&ipv4address[i]->ipv4_address, G_SOCKET_FAMILY_IPV4);
                 str = g_inet_address_to_string (addr);
-                mm_dbg ("    IP [%u]: '%s/%u'", i, str, ipv4address[i]->on_link_prefix_length);
+                mm_obj_dbg (self, "    IP [%u]: '%s/%u'", i, str, ipv4address[i]->on_link_prefix_length);
                 g_free (str);
                 g_object_unref (addr);
             }
@@ -320,7 +323,7 @@ ip_configuration_query_ready (MbimDevice *device,
         if ((ipv4configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_GATEWAY) && ipv4gateway) {
             addr = g_inet_address_new_from_bytes ((guint8 *)ipv4gateway, G_SOCKET_FAMILY_IPV4);
             str = g_inet_address_to_string (addr);
-            mm_dbg ("  Gateway: '%s'", str);
+            mm_obj_dbg (self, "  gateway: '%s'", str);
             g_free (str);
             g_object_unref (addr);
         }
@@ -328,12 +331,12 @@ ip_configuration_query_ready (MbimDevice *device,
         if ((ipv4configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_DNS) && ipv4dnsservercount) {
             guint i;
 
-            mm_dbg ("  DNS addresses (%u)", ipv4dnsservercount);
+            mm_obj_dbg (self, "  DNS addresses (%u)", ipv4dnsservercount);
             for (i = 0; i < ipv4dnsservercount; i++) {
                 addr = g_inet_address_new_from_bytes ((guint8 *)&ipv4dnsserver[i], G_SOCKET_FAMILY_IPV4);
                 if (!g_inet_address_get_is_any (addr)) {
                     str = g_inet_address_to_string (addr);
-                    mm_dbg ("    DNS [%u]: '%s'", i, str);
+                    mm_obj_dbg (self, "    DNS [%u]: '%s'", i, str);
                     g_free (str);
                 }
                 g_object_unref (addr);
@@ -341,23 +344,23 @@ ip_configuration_query_ready (MbimDevice *device,
         }
 
         if ((ipv4configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_MTU) && ipv4mtu) {
-            mm_dbg ("  MTU: '%u'", ipv4mtu);
+            mm_obj_dbg (self, "  MTU: '%u'", ipv4mtu);
         }
 
         /* IPv6 info */
 
         str = mbim_ip_configuration_available_flag_build_string_from_mask (ipv6configurationavailable);
-        mm_dbg ("IPv6 configuration available: '%s'", str);
+        mm_obj_dbg (self, "IPv6 configuration available: '%s'", str);
         g_free (str);
 
         if ((ipv6configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_ADDRESS) && ipv6addresscount) {
             guint i;
 
-            mm_dbg ("  IP addresses (%u)", ipv6addresscount);
+            mm_obj_dbg (self, "  IP addresses (%u)", ipv6addresscount);
             for (i = 0; i < ipv6addresscount; i++) {
                 addr = g_inet_address_new_from_bytes ((guint8 *)&ipv6address[i]->ipv6_address, G_SOCKET_FAMILY_IPV6);
                 str = g_inet_address_to_string (addr);
-                mm_dbg ("    IP [%u]: '%s/%u'", i, str, ipv6address[i]->on_link_prefix_length);
+                mm_obj_dbg (self, "    IP [%u]: '%s/%u'", i, str, ipv6address[i]->on_link_prefix_length);
                 g_free (str);
                 g_object_unref (addr);
             }
@@ -366,7 +369,7 @@ ip_configuration_query_ready (MbimDevice *device,
         if ((ipv6configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_GATEWAY) && ipv6gateway) {
             addr = g_inet_address_new_from_bytes ((guint8 *)ipv6gateway, G_SOCKET_FAMILY_IPV6);
             str = g_inet_address_to_string (addr);
-            mm_dbg ("  Gateway: '%s'", str);
+            mm_obj_dbg (self, "  gateway: '%s'", str);
             g_free (str);
             g_object_unref (addr);
         }
@@ -374,12 +377,12 @@ ip_configuration_query_ready (MbimDevice *device,
         if ((ipv6configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_DNS) && ipv6dnsservercount) {
             guint i;
 
-            mm_dbg ("  DNS addresses (%u)", ipv6dnsservercount);
+            mm_obj_dbg (self, "  DNS addresses (%u)", ipv6dnsservercount);
             for (i = 0; i < ipv6dnsservercount; i++) {
                 addr = g_inet_address_new_from_bytes ((guint8 *)&ipv6dnsserver[i], G_SOCKET_FAMILY_IPV6);
                 if (!g_inet_address_get_is_any (addr)) {
                     str = g_inet_address_to_string (addr);
-                    mm_dbg ("    DNS [%u]: '%s'", i, str);
+                    mm_obj_dbg (self, "    DNS [%u]: '%s'", i, str);
                     g_free (str);
                 }
                 g_object_unref (addr);
@@ -387,7 +390,7 @@ ip_configuration_query_ready (MbimDevice *device,
         }
 
         if ((ipv6configurationavailable & MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_MTU) && ipv6mtu) {
-            mm_dbg ("  MTU: '%u'", ipv6mtu);
+            mm_obj_dbg (self, "  MTU: '%u'", ipv6mtu);
         }
 
         /* Build connection results */
@@ -562,14 +565,17 @@ connect_set_ready (MbimDevice *device,
                    GAsyncResult *res,
                    GTask *task)
 {
+    MMBearerMbim *self;
     ConnectContext *ctx;
     GError *error = NULL;
     MbimMessage *response;
     guint32 session_id;
+    MbimContextIpType ip_type;
     MbimActivationState activation_state;
     guint32 nw_error;
 
-    ctx = g_task_get_task_data (task);
+    self = g_task_get_source_object (task);
+    ctx  = g_task_get_task_data (task);
 
     response = mbim_device_command_finish (device, res, &error);
     if (response &&
@@ -582,25 +588,36 @@ connect_set_ready (MbimDevice *device,
                 &session_id,
                 &activation_state,
                 NULL, /* voice_call_state */
-                NULL, /* ip_type */
+                &ip_type,
                 NULL, /* context_type */
                 &nw_error,
                 &inner_error)) {
-            if (nw_error) {
-                if (error)
-                    g_error_free (error);
-                error = mm_mobile_equipment_error_from_mbim_nw_error (nw_error);
-            } else {
-                /* Report the ip_type we originally requested, since the ip_type
-                 * from the response is only relevant if the requested used
-                 * MBIM_CONTEXT_IP_TYPE_DEFAULT, which MM never does.  Some
-                 * devices (K5160) report the wrong type in the response.
-                 */
-                mm_dbg ("Session ID '%u': %s (IP type: %s)",
+            /* Report the IP type we asked for and the one returned by the modem */
+            mm_obj_dbg (self, "session ID '%u': %s (requested IP type: %s, received IP type: %s, nw error: %s)",
                         session_id,
                         mbim_activation_state_get_string (activation_state),
-                        mbim_context_ip_type_get_string (ctx->ip_type));
+                        mbim_context_ip_type_get_string (ctx->ip_type),
+                        mbim_context_ip_type_get_string (ip_type),
+                        nw_error ? mbim_nw_error_get_string (nw_error) : "none");
+            /* If the response reports an ACTIVATED state, we're good even if
+             * there is a nw_error set (e.g. asking for IPv4v6 may return a
+             * 'pdp-type-ipv4-only-allowed' nw_error). */
+            if (activation_state != MBIM_ACTIVATION_STATE_ACTIVATED &&
+                activation_state != MBIM_ACTIVATION_STATE_ACTIVATING) {
+                if (nw_error) {
+                    g_clear_error (&error);
+                    error = mm_mobile_equipment_error_from_mbim_nw_error (nw_error);
+                } else if (!error) {
+                    error = g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
+                                         MM_MOBILE_EQUIPMENT_ERROR_GPRS_UNKNOWN,
+                                         "Unknown error: context activation failed");
+                }
             }
+            /* We're now connected, but we may have received an IP type different to the one
+             * requested (e.g. asking for IPv4v6 but received IPv4 only). Handle that, so that
+             * the next step getting IP details doesn't get confused. */
+            else if (ctx->ip_type != ip_type)
+                ctx->ip_type = ip_type;
         } else {
             /* Prefer the error from the result to the parsing error */
             if (!error)
@@ -644,18 +661,69 @@ ensure_disconnected_ready (MbimDevice   *device,
     connect_context_step (task);
 }
 
+static void
+check_disconnected_ready (MbimDevice   *device,
+                          GAsyncResult *res,
+                          GTask        *task)
+{
+    MMBearerMbim *self;
+    ConnectContext *ctx;
+    GError *error = NULL;
+    MbimMessage *response;
+    guint32 session_id;
+    MbimActivationState activation_state;
+
+    self = g_task_get_source_object (task);
+    ctx  = g_task_get_task_data (task);
+
+    response = mbim_device_command_finish (device, res, &error);
+    if (response &&
+        mbim_message_response_get_result (response, MBIM_MESSAGE_TYPE_COMMAND_DONE, &error) &&
+        mbim_message_connect_response_parse (
+            response,
+            &session_id,
+            &activation_state,
+            NULL, /* voice_call_state */
+            NULL, /* ip_type */
+            NULL, /* context_type */
+            NULL, /* nw_error */
+            &error)) {
+        mm_obj_dbg (self, "session ID '%u': %s", session_id, mbim_activation_state_get_string (activation_state));
+    } else
+        activation_state = MBIM_ACTIVATION_STATE_UNKNOWN;
+
+    if (response)
+        mbim_message_unref (response);
+
+    /* Some modem (e.g. Huawei ME936) reports MBIM_ACTIVATION_STATE_UNKNOWN
+     * when being queried for the activation state before an IP session has
+     * been activated once. Here we expect a modem would at least tell the
+     * truth when the session has been activated, so we proceed to deactivate
+     * the session only the modem indicates the session has been activated or
+     * is being activated.
+     */
+    if (activation_state == MBIM_ACTIVATION_STATE_ACTIVATED || activation_state == MBIM_ACTIVATION_STATE_ACTIVATING)
+        ctx->step = CONNECT_STEP_ENSURE_DISCONNECTED;
+    else
+        ctx->step = CONNECT_STEP_CONNECT;
+
+    connect_context_step (task);
+}
+
 static void
 provisioned_contexts_query_ready (MbimDevice *device,
                                   GAsyncResult *res,
                                   GTask *task)
 {
+    MMBearerMbim *self;
     ConnectContext *ctx;
     GError *error = NULL;
     MbimMessage *response;
     guint32 provisioned_contexts_count;
     MbimProvisionedContextElement **provisioned_contexts;
 
-    ctx = g_task_get_task_data (task);
+    self = g_task_get_source_object (task);
+    ctx  = g_task_get_task_data (task);
 
     response = mbim_device_command_finish (device, res, &error);
     if (response &&
@@ -667,25 +735,25 @@ provisioned_contexts_query_ready (MbimDevice *device,
             &error)) {
         guint32 i;
 
-        mm_dbg ("Provisioned contexts found (%u):", provisioned_contexts_count);
+        mm_obj_dbg (self, "provisioned contexts found (%u):", provisioned_contexts_count);
         for (i = 0; i < provisioned_contexts_count; i++) {
             MbimProvisionedContextElement *el = provisioned_contexts[i];
             gchar *uuid_str;
 
             uuid_str = mbim_uuid_get_printable (&el->context_type);
-            mm_dbg ("[%u] context type: %s", el->context_id, mbim_context_type_get_string (mbim_uuid_to_context_type (&el->context_type)));
-            mm_dbg ("             uuid: %s", uuid_str);
-            mm_dbg ("    access string: %s", el->access_string ? el->access_string : "");
-            mm_dbg ("         username: %s", el->user_name ? el->user_name : "");
-            mm_dbg ("         password: %s", el->password ? el->password : "");
-            mm_dbg ("      compression: %s", mbim_compression_get_string (el->compression));
-            mm_dbg ("             auth: %s", mbim_auth_protocol_get_string (el->auth_protocol));
+            mm_obj_dbg (self, "[%u] context type: %s", el->context_id, mbim_context_type_get_string (mbim_uuid_to_context_type (&el->context_type)));
+            mm_obj_dbg (self, "             uuid: %s", uuid_str);
+            mm_obj_dbg (self, "    access string: %s", el->access_string ? el->access_string : "");
+            mm_obj_dbg (self, "         username: %s", el->user_name ? el->user_name : "");
+            mm_obj_dbg (self, "         password: %s", el->password ? el->password : "");
+            mm_obj_dbg (self, "      compression: %s", mbim_compression_get_string (el->compression));
+            mm_obj_dbg (self, "             auth: %s", mbim_auth_protocol_get_string (el->auth_protocol));
             g_free (uuid_str);
         }
 
         mbim_provisioned_context_element_array_free (provisioned_contexts);
     } else {
-        mm_dbg ("Error listing provisioned contexts: %s", error->message);
+        mm_obj_dbg (self, "error listing provisioned contexts: %s", error->message);
         g_error_free (error);
     }
 
@@ -702,6 +770,7 @@ packet_service_set_ready (MbimDevice *device,
                           GAsyncResult *res,
                           GTask *task)
 {
+    MMBearerMbim *self;
     ConnectContext *ctx;
     GError *error = NULL;
     MbimMessage *response;
@@ -711,6 +780,7 @@ packet_service_set_ready (MbimDevice *device,
     guint64 uplink_speed;
     guint64 downlink_speed;
 
+    self = g_task_get_source_object (task);
     ctx = g_task_get_task_data (task);
 
     response = mbim_device_command_finish (device, res, &error);
@@ -735,11 +805,11 @@ packet_service_set_ready (MbimDevice *device,
                 gchar *str;
 
                 str = mbim_data_class_build_string_from_mask (highest_available_data_class);
-                mm_dbg ("Packet service update:");
-                mm_dbg ("         state: '%s'", mbim_packet_service_state_get_string (packet_service_state));
-                mm_dbg ("    data class: '%s'", str);
-                mm_dbg ("        uplink: '%" G_GUINT64_FORMAT "' bps", uplink_speed);
-                mm_dbg ("      downlink: '%" G_GUINT64_FORMAT "' bps", downlink_speed);
+                mm_obj_dbg (self, "packet service update:");
+                mm_obj_dbg (self, "         state: '%s'", mbim_packet_service_state_get_string (packet_service_state));
+                mm_obj_dbg (self, "    data class: '%s'", str);
+                mm_obj_dbg (self, "        uplink: '%" G_GUINT64_FORMAT "' bps", uplink_speed);
+                mm_obj_dbg (self, "      downlink: '%" G_GUINT64_FORMAT "' bps", downlink_speed);
                 g_free (str);
             }
         } else {
@@ -758,7 +828,7 @@ packet_service_set_ready (MbimDevice *device,
         /* Don't make NoDeviceSupport errors fatal; just try to keep on the
          * connection sequence even with this error. */
         if (g_error_matches (error, MBIM_STATUS_ERROR, MBIM_STATUS_ERROR_NO_DEVICE_SUPPORT)) {
-            mm_dbg ("Device doesn't support packet service attach");
+            mm_obj_dbg (self, "device doesn't support packet service attach");
             g_error_free (error);
         } else {
             /* All other errors are fatal */
@@ -791,13 +861,13 @@ connect_context_step (GTask *task)
 
     switch (ctx->step) {
     case CONNECT_STEP_FIRST:
-        /* Fall down */
         ctx->step++;
+        /* Fall through */
 
     case CONNECT_STEP_PACKET_SERVICE: {
         GError *error = NULL;
 
-        mm_dbg ("Activating packet service...");
+        mm_obj_dbg (self, "activating packet service...");
         message = (mbim_message_packet_service_set_new (
                        MBIM_PACKET_SERVICE_ACTION_ATTACH,
                        &error));
@@ -818,7 +888,7 @@ connect_context_step (GTask *task)
     }
 
     case CONNECT_STEP_PROVISIONED_CONTEXTS:
-        mm_dbg ("Listing provisioned contexts...");
+        mm_obj_dbg (self, "listing provisioned contexts...");
         message = mbim_message_provisioned_contexts_query_new (NULL);
         mbim_device_command (ctx->device,
                              message,
@@ -829,8 +899,34 @@ connect_context_step (GTask *task)
         mbim_message_unref (message);
         return;
 
+    case CONNECT_STEP_CHECK_DISCONNECTED: {
+        GError *error = NULL;
+
+        message = (mbim_message_connect_query_new (
+                       self->priv->session_id,
+                       MBIM_ACTIVATION_STATE_UNKNOWN,
+                       MBIM_VOICE_CALL_STATE_NONE,
+                       MBIM_CONTEXT_IP_TYPE_DEFAULT,
+                       mbim_uuid_from_context_type (MBIM_CONTEXT_TYPE_INTERNET),
+                       0,
+                       &error));
+        if (!message) {
+            g_task_return_error (task, error);
+            g_object_unref (task);
+            return;
+        }
+
+        mbim_device_command (ctx->device,
+                             message,
+                             10,
+                             NULL,
+                             (GAsyncReadyCallback)check_disconnected_ready,
+                             task);
+        mbim_message_unref (message);
+        return;
+    }
+
     case CONNECT_STEP_ENSURE_DISCONNECTED: {
-        MbimMessage *message;
         GError *error = NULL;
 
         message = (mbim_message_connect_set_new (
@@ -852,7 +948,7 @@ connect_context_step (GTask *task)
 
         mbim_device_command (ctx->device,
                              message,
-                             30,
+                             60,
                              NULL,
                              (GAsyncReadyCallback)ensure_disconnected_ready,
                              task);
@@ -878,30 +974,12 @@ connect_context_step (GTask *task)
             auth = MBIM_AUTH_PROTOCOL_NONE;
         } else {
             MMBearerAllowedAuth bearer_auth;
-            bearer_auth = mm_bearer_properties_get_allowed_auth (ctx->properties);
-            if (bearer_auth == MM_BEARER_ALLOWED_AUTH_UNKNOWN) {
-                mm_dbg ("Using default (PAP) authentication method");
-                auth = MBIM_AUTH_PROTOCOL_PAP;
-            } else if (bearer_auth & MM_BEARER_ALLOWED_AUTH_PAP) {
-                auth = MBIM_AUTH_PROTOCOL_PAP;
-            } else if (bearer_auth & MM_BEARER_ALLOWED_AUTH_CHAP) {
-                auth = MBIM_AUTH_PROTOCOL_CHAP;
-            } else if (bearer_auth & MM_BEARER_ALLOWED_AUTH_MSCHAPV2) {
-                auth = MBIM_AUTH_PROTOCOL_MSCHAPV2;
-            } else if (bearer_auth & MM_BEARER_ALLOWED_AUTH_NONE) {
-                auth = MBIM_AUTH_PROTOCOL_NONE;
-            } else {
-                gchar *str;
 
-                str = mm_bearer_allowed_auth_build_string_from_mask (bearer_auth);
-                g_task_return_new_error (
-                    task,
-                    MM_CORE_ERROR,
-                    MM_CORE_ERROR_UNSUPPORTED,
-                    "Cannot use any of the specified authentication methods (%s)",
-                    str);
+            bearer_auth = mm_bearer_properties_get_allowed_auth (ctx->properties);
+            auth = mm_bearer_allowed_auth_to_mbim_auth_protocol (bearer_auth, self, &error);
+            if (error) {
+                g_task_return_error (task, error);
                 g_object_unref (task);
-                g_free (str);
                 return;
             }
         }
@@ -913,38 +991,18 @@ connect_context_step (GTask *task)
 
             ip_family = mm_base_bearer_get_default_ip_family (MM_BASE_BEARER (self));
             str = mm_bearer_ip_family_build_string_from_mask (ip_family);
-            mm_dbg ("No specific IP family requested, defaulting to %s", str);
+            mm_obj_dbg (self, "no specific IP family requested, defaulting to %s", str);
             g_free (str);
         }
 
-        if (ip_family == MM_BEARER_IP_FAMILY_IPV4)
-            ctx->ip_type = MBIM_CONTEXT_IP_TYPE_IPV4;
-        else if (ip_family == MM_BEARER_IP_FAMILY_IPV6)
-            ctx->ip_type = MBIM_CONTEXT_IP_TYPE_IPV6;
-        else if (ip_family == MM_BEARER_IP_FAMILY_IPV4V6)
-            ctx->ip_type = MBIM_CONTEXT_IP_TYPE_IPV4V6;
-        else if (ip_family == (MM_BEARER_IP_FAMILY_IPV4 | MM_BEARER_IP_FAMILY_IPV6))
-            ctx->ip_type = MBIM_CONTEXT_IP_TYPE_IPV4_AND_IPV6;
-        else if (ip_family == MM_BEARER_IP_FAMILY_NONE ||
-                 ip_family == MM_BEARER_IP_FAMILY_ANY)
-            /* A valid default IP family should have been specified */
-            g_assert_not_reached ();
-        else  {
-            gchar * str;
-
-            str = mm_bearer_ip_family_build_string_from_mask (ip_family);
-            g_task_return_new_error (
-                task,
-                MM_CORE_ERROR,
-                MM_CORE_ERROR_UNSUPPORTED,
-                "Unsupported IP type configuration: '%s'",
-                str);
+        ctx->ip_type = mm_bearer_ip_family_to_mbim_context_ip_type (ip_family, &error);
+        if (error) {
+            g_task_return_error (task, error);
             g_object_unref (task);
-            g_free (str);
             return;
         }
 
-        mm_dbg ("Launching %s connection with APN '%s'...", mbim_context_ip_type_get_string (ctx->ip_type), apn);
+        mm_obj_dbg (self, "launching %s connection with APN '%s'...", mbim_context_ip_type_get_string (ctx->ip_type), apn);
         message = (mbim_message_connect_set_new (
                        self->priv->session_id,
                        MBIM_ACTIVATION_COMMAND_ACTIVATE,
@@ -975,7 +1033,7 @@ connect_context_step (GTask *task)
     case CONNECT_STEP_IP_CONFIGURATION: {
         GError *error = NULL;
 
-        mm_dbg ("Querying IP configuration...");
+        mm_obj_dbg (self, "querying IP configuration...");
         message = (mbim_message_ip_configuration_query_new (
                        self->priv->session_id,
                        MBIM_IP_CONFIGURATION_AVAILABLE_FLAG_NONE, /* ipv4configurationavailable */
@@ -1024,6 +1082,9 @@ connect_context_step (GTask *task)
             (GDestroyNotify)mm_bearer_connect_result_unref);
         g_object_unref (task);
         return;
+
+    default:
+        break;
     }
 
     g_assert_not_reached ();
@@ -1069,7 +1130,7 @@ _connect (MMBaseBearer *self,
 
     g_object_unref (modem);
 
-    mm_dbg ("Launching connection with data port (%s/%s)",
+    mm_obj_dbg (self, "launching connection with data port (%s/%s)",
             mm_port_subsys_get_string (mm_port_get_subsys (data)),
             mm_port_get_device (data));
 
@@ -1135,6 +1196,7 @@ disconnect_set_ready (MbimDevice *device,
                       GAsyncResult *res,
                       GTask *task)
 {
+    MMBearerMbim *self;
     DisconnectContext *ctx;
     GError *error = NULL;
     MbimMessage *response;
@@ -1144,7 +1206,8 @@ disconnect_set_ready (MbimDevice *device,
     GError *inner_error = NULL;
     gboolean result = FALSE, parsed_result = FALSE;
 
-    ctx = g_task_get_task_data (task);
+    self = g_task_get_source_object (task);
+    ctx  = g_task_get_task_data (task);
 
     response = mbim_device_command_finish (device, res, &error);
     if (!response)
@@ -1172,16 +1235,16 @@ disconnect_set_ready (MbimDevice *device,
     if (result && parsed_result) {
         g_assert (!error);
         g_assert (!inner_error);
-        mm_dbg ("Session ID '%u': %s", session_id, mbim_activation_state_get_string (activation_state));
+        mm_obj_dbg (self, "session ID '%u': %s", session_id, mbim_activation_state_get_string (activation_state));
         /* success */
         goto out;
     }
 
     if (g_error_matches (error, MBIM_STATUS_ERROR, MBIM_STATUS_ERROR_CONTEXT_NOT_ACTIVATED)) {
         if (parsed_result)
-            mm_dbg ("Context not activated: session ID '%u' already disconnected", session_id);
+            mm_obj_dbg (self, "context not activated: session ID '%u' already disconnected", session_id);
         else
-            mm_dbg ("Context not activated: already disconnected");
+            mm_obj_dbg (self, "context not activated: already disconnected");
 
         g_clear_error (&error);
         g_clear_error (&inner_error);
@@ -1229,8 +1292,8 @@ disconnect_context_step (GTask *task)
 
     switch (ctx->step) {
     case DISCONNECT_STEP_FIRST:
-        /* Fall down */
         ctx->step++;
+        /* Fall through */
 
     case DISCONNECT_STEP_DISCONNECT: {
         MbimMessage *message;
@@ -1255,7 +1318,7 @@ disconnect_context_step (GTask *task)
 
         mbim_device_command (ctx->device,
                              message,
-                             30,
+                             60,
                              NULL,
                              (GAsyncReadyCallback)disconnect_set_ready,
                              task);
@@ -1270,6 +1333,9 @@ disconnect_context_step (GTask *task)
         g_task_return_boolean (task, TRUE);
         g_object_unref (task);
         return;
+
+    default:
+        g_assert_not_reached ();
     }
 }
 
@@ -1283,31 +1349,27 @@ disconnect (MMBaseBearer *_self,
     DisconnectContext *ctx;
     GTask *task;
 
-    if (!self->priv->data) {
-        g_task_report_new_error (
-            self,
-            callback,
-            user_data,
-            disconnect,
-            MM_CORE_ERROR,
-            MM_CORE_ERROR_FAILED,
-            "Couldn't disconnect MBIM bearer: this bearer is not connected");
-        return;
-    }
-
     if (!peek_ports (self, &device, NULL, callback, user_data))
         return;
 
-    mm_dbg ("Launching disconnection on data port (%s/%s)",
-            mm_port_subsys_get_string (mm_port_get_subsys (self->priv->data)),
-            mm_port_get_device (self->priv->data));
+    task = g_task_new (self, NULL, callback, user_data);
+
+    if (!self->priv->data) {
+        mm_obj_dbg (self, "no need to disconnect: MBIM bearer is already disconnected");
+        g_task_return_boolean (task, TRUE);
+        g_object_unref (task);
+        return;
+    }
+
+    mm_obj_dbg (self, "launching disconnection on data port (%s/%s)",
+                mm_port_subsys_get_string (mm_port_get_subsys (self->priv->data)),
+                mm_port_get_device (self->priv->data));
 
     ctx = g_slice_new0 (DisconnectContext);
     ctx->device = g_object_ref (device);
     ctx->data = g_object_ref (self->priv->data);
     ctx->step = DISCONNECT_STEP_FIRST;
 
-    task = g_task_new (self, NULL, callback, user_data);
     g_task_set_task_data (task, ctx, (GDestroyNotify)disconnect_context_free);
 
     /* Run! */
